# File containing the classes corresponding to each construction of the
# simple syntax of RELAX NG
require 'callbacks'
module Srnggen
  class Document
    def def_class
      Node::Define
    end
    def root_class
      Node::Start
    end
  end

  class Node
    #class representing a RELAX NG grammar.
    class Grammar < Node
      include Comb::Grammar
      def initialize xml, doc
        super
        xml.each_element {|e| @elems << Node.parse(e, doc)}
      end

      def gen
        elem.gen
      end
    end

    class Start < Node
      include Comb::Marker
      def name; "start" end
      def gen
        elem.gen
      end
    end

    class Define < Node
      # each element contributes 1 to the size. A definition may concern a set
      # of names or a class of names, we must thus take that in account for the
      # weight of the definition.
      include Comb::Def
      include Comb::Marker
      def initialize xml, doc
        super
        @name = xml.attributes["name"]
      end
      attr :name
      def gen
        elem.gen
      end
    end

    class Element < Node
      include Comb::Product
      def initialize xml, doc
        super
        @elems << Comb::Mark.new(doc)
      end

      def gen
        Res.new [elem(2).gen.to_e(elem(1).gen.to_s)]
      end
    end

    class Notallowed < Node
      include Comb::Epsilon
      def initialize xml, doc
        super nil, doc
      end
      #TODO
    end

    class Except < Node
      include Comb::Epsilon
      def initialize xml, doc
        super nil, doc
      end
      #TODO
    end

    class Param < Node
      include Comb::Epsilon
      def initialize xml, doc
        super nil, doc
      end
      #TODO
    end

    class Empty < Node
      include Comb::Epsilon
      def gen
        Res.new
      end
    end

    class Text < Node
      #TODO parametrer la longueur du texte produit
      #TODO donner le choix de peser x ou len*x
      include Comb::Atom

      def weight
        Callbacks.weight_text @xml
      end

      def gen
        Res.new [Callbacks.gen_text(@xml)]
      end
    end

    class Data < Node
      include Comb::Atom
      def initialize xml, doc
        super
        @type = xml.attributes["type"]
        @lib  = xml.attributes["datatypeLibrary"]
      end

      def weight
        Callbacks.weight(@type, @xml)
      end

      def gen
        Res.new [Callbacks.gen(@type, @xml)]
      end
    end

    class Value < Node
      include Comb::Atom
      def initialize xml, doc
        super
#        @lib  = xml.attributes["datatypeLibrary"]
#        @type = xml.attributes["type"]
#        @ns   = xml.attributes["ns"]
        @val  = xml.text
      end

      def gen
        Res.new [@val]
      end
    end

    class List < Node
      include Comb::Marker
      # res should only contain strings
      def gen
        Res.new [elem.gen.compact]
      end
    end

    class Attribute < Node
      include Comb::Product
      def initialize xml, doc
        super
        @elems << Comb::Mark.new(doc)
      end

      def gen
        Res.new [], {elem(1).gen.to_s => elem(2).gen.to_s}
      end
    end

    class Ref < Node
      include Comb::Reference
      attr_reader :ref
      def initialize xml, doc
        super
        @ref = xml.attributes["name"]
      end

      def gen
        @doc.defs[@ref].gen
      end
    end

    class Oneormore < Node
      include Comb::Sequence
      def gen
        r = Res.new
        Sequence.geom(@magic).times {r.join elem.gen}
        r
      end
    end

    class Choice < Node
      include Comb::Union
      def gen
        elem(rand < @magic[0] ? 1 : 2).gen
      end
    end

    class Group < Node
      include Comb::Product
      def gen
        elem(1).gen.join elem(2).gen
      end
    end

    #TODO this is not really a product but a union of the attributes
    #and an interleaving of the elements of the children
    class Interleave < Node
      include Comb::Product
      def gen
        elem(1).gen.interleave elem(2).gen
      end
    end

    class Anyname < Node
      include Comb::Epsilon
      def gen
        Res.new ["anyname"]
      end
    end

    class Nsname < Node
      include Comb::Epsilon
      def initialize xml, doc
        super
        @ns = xml.attributes["ns"]
      end

      def gen
        Res.new ["nsname"]
      end
    end

    class Name < Node
      include Comb::Epsilon
      def initialize xml, doc
        super
#        @ns = xml.attributes["ns"]
        @name = xml.text
      end

      def gen
        Res.new [@name]
      end
    end
  end
end
