#!/usr/bin/ruby

# :main: srnggen.rb
# 
# = SrngGen - Simple RELAX NG Generation
# 
# This package contains a simple and flexible sampler for XML documents respecting
# a given
# {RELAX NG}[http://www.oasis-open.org/committees/relax-ng/spec-20011203.html]
# grammar. It can be used as a standalone application or as a
# library.
# 
# The RELAX NG grammar given must use the
# {simple syntax}[http://www.oasis-open.org/committees/relax-ng/spec-20011203.html#simple-syntax].
# You can transform any RELAX NG grammar into its equivalent simplified syntax
# using the rng2srng[http://www.kohsuke.org/relaxng/rng2srng/] tool.
# 
# To calculate values needed for the sampling Maple[http://www.maplesoft.com/] is
# used as a co-routine. The program thus expects that <tt>maple</tt> is in the
# path. There is however a way, explained further, to invoke Maple manually.
# 
# == Overview
# 
# This package is intended as a prototype to demonstrate how we can easily apply
# recent developments in random sampling provided by combinatory theory to
# generate random inputs of good quality for testing computer programs.
# 
# In this prototype we focus on programs that read XML documents respecting a
# given RELAX NG grammar. A RELAX NG grammar can easily be translated to the
# combinatorial description of a class of trees. From there we can automatically
# derive a random sampler for objects in that class. 
# 
# == Summary
# 
# You can generate 10 random XML documents respecting a given grammar
#   % ruby srnggen.rb -n 10 grammar.srng
# The resulting XML will be placed in the files <tt>result*.xml.</tt>
# 
# If you plan to sample many times using the same grammar, you can keep the result
# of the Maple invocation in a file
#   % ruby srnggen.rb -c grammar.srng > oracle
# and later use this file for the sampling
#   % ruby srnggen.rb -O oracle grammar.srng
# 
# If Maple is not installed on your machine, you can get the Maple program in a
# file to execute it elsewhere
#   % ruby srnggen.rb -m grammar.srng > program.mpl
#   % maple -q program.mpl > oracle
# 
# == Usage
# 
# srnggen.rb [OPTION] ... grammar.srng [oracle]
# 
# Options are:
# 
# [<tt>-h, --help</tt>]
#    show help
# 
# [<tt>--compile, -c</tt>]
#    calculate oracle values
# 
# [<tt>--n</tt> <i>N</i>]
#    number of xml documents to produce
# 
# [<tt>--oracle, -O</tt>]
#    file containing oracle values
# 
# [<tt>--maple, -m</tt>]
#    output maple programm to calculate oracle values
# 
# [<tt>--precision</tt> <i>N</i>, <tt>-p</tt> <i>N</i>]
#    precision for the calculation of the singularity
# 
# [<tt>--callbacks</tt> <i>File</i>, <tt>-C</tt> <i>File</i>]
#    file containing callbacks definitions

require 'rexml/document'
require 'tmpdir'
require 'node'
require 'comb'
require 'res'
require 'oracle'

module Srnggen
  class Document
    def initialize source
      @defs, @oracle = {}, {}
      @doc = REXML::Document.new source
      @node = Node.parse @doc.root, self
      @node.elems.each do |e|
        case e
        when def_class:
          @defs[e.name] = e
        when root_class:
          @root = e
        end
      end
      @eq = @defs.keys
    end

    # Definition objects for all the elements except root
    attr :defs
    attr :root
    # The Oracle object containing the values needed for the sampling
    attr :oracle

    # generate one random XML document
    def gen
      @node.gen.element
    end

    # use the values in an Oracle to calculate all the intermediate values
    # needed for the sampling
    def compile oracle
      @oracle = oracle
      @node.compile
    end

    def inspect
      "<#{self.class} #{defs.size + 1} definitions>"
    end

    def simplify
      h, eq = {}, []
      # group definitions by value
      @defs.each_pair do |k,v|
        e = v.to_sys
        h[e] = [] unless h.has_key? e
        h[e] << k
      end
      # keep one definitions for each value
      h.each_value do |a|
        t = a.shift
        eq << t
        a.each do |n|
          @defs[n].extend Comb::Aliased
          @defs[n].set_content t
          @node.elems.each {|e| e.replace n, t}
        end
      end
      # remove definitions with no indets in value
      tmp = eq.collect do |n|
        n if @defs[n].indets(Hash.new).keys.empty?
      end.compact
      tmp.each do |n|
        eq.each {|v| @defs[v].replace n, @defs[n]}
        root.replace n, @defs[n]
        eq.delete tmp
      end
      # remove non-recursive definitions
      loop do
        tmp = eq.collect do |n|
          n unless @defs[n].indets(Hash.new).keys.include? n
        end.compact.first
        break unless tmp
        eq.each {|v| @defs[v].replace tmp, @defs[tmp]}
        root.replace tmp, @defs[tmp]
        eq.delete tmp
      end
      @eq = eq
    end
  end
end

if __FILE__ == $0
  require 'getoptlong'

  opts = GetoptLong.new(
    [ '--help',       '-h', GetoptLong::NO_ARGUMENT ],
    [ '--compile',    '-c', GetoptLong::NO_ARGUMENT ],
    [ '--system',     '-s', GetoptLong::NO_ARGUMENT ],
    [ '--backend',    '-b', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--language',   '-l', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--oracle',     '-O', GetoptLong::REQUIRED_ARGUMENT ],
    [                 '-n', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--precision',  '-p', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--callbacks',  '-C', GetoptLong::REQUIRED_ARGUMENT ],
    [ '--no-simplify',      GetoptLong::NO_ARGUMENT ]
  )
  compile, system, backend,  language, n, oracle_file, precision,   d, callbacks, nosimplify =
    false,  false, "c",     "relaxng", 1,         nil,       "5", nil,       nil,      false
  opts.each do |opt, arg|
    case opt
    when '--help':        RDoc::usage
    when '--compile':     compile = true
    when '--system':      system = true
    when '--backend':     backend = arg
    when '--language':    language = arg
    when '-n':            n = arg.to_i
    when '--oracle':      oracle_file = arg
    when '--precision':   precision = arg
    when '--callbacks':   callbacks = arg
    when '--no-simplify': nosimplify = true
    end
  end
  if ARGV.length != 1 then
    require 'rdoc/usage'
    RDoc::usage(0, 'Usage')
  end
  require "backend/" + backend
  require "language/" + language
  File.open(ARGV.shift) {|f| d = Srnggen::Document.new f}
  Srnggen.module_eval IO.readlines(callbacks).join if callbacks
  d.simplify unless nosimplify or oracle_file
  if system then d.to_sys $stdout; exit 0 end
  oracle = Srnggen::Oracle.new(oracle_file ? oracle_file : d)
  if compile then puts oracle; exit 0 end
  d.compile oracle
  writer = REXML::Formatters::Pretty.new
  writer.width = 8096
  1.upto n do |i|
    File.open("result#{i}.xml", "w") {|f| writer.write d.gen, f; f.puts}
  end
end
